/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: Astro.c,v 1.19 2000/09/27 20:36:56 mheinz Exp $
 * Copyright (C) 2000, Michael Heinz
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include <PalmOS.h>
#include "PalmUtil.h"
#include "resource.h"
#include "MathLib.h"
#include "Astro.h"
#include "AstroLib.h"
#include "AstroCat.h"
#include "constants.h"

SystemPreferencesType   sysPrefs;
MoonPrefsType           moonPrefs;
UInt16                  mptSize = sizeof(MoonPrefsType);

/*
 * Pointers to various timer controls
 */
static ControlPtr      datePtr, timePtr;
static ControlPtr      date2Ptr, time2Ptr;
static ControlPtr      date3Ptr, time3Ptr;
static ControlPtr      date4Ptr, time4Ptr;

/*
 * Pointers to phase form objects
 */
static MemHandle       bitmapHandle;
static RectangleType   bar  = {{110, 81}, {0, 8}};


/*
 * This is used to remember the previous form when a dialog is up.
 * It's not needed generally, but we use it so moonPrefs doesn't get
 * out of sync.
 */
UInt16                 saveFormId;

/*
 * Does this palm support Color or only black and white? Note that we
 * use a different menu for color palms.
 */
static Boolean         colorSupport = false;
UInt16                 mainMenu = m_Normal;

/*
 * Update a text field and redraw it.
 */
void
UpdateTextField(FieldPtr fld, char *t)
{
    char        *s; 
    MemHandle   h;
    UInt16      size, length;
    
    if (fld)
    {
        /*
         * Make sure the text will fit.
         */
        size = FldGetMaxChars(fld);
        length = StrLen(t);
        if (length>size)
            length = size;
        
        FldFreeMemory(fld);
        h = MemHandleNew(length+1);

        if (h) {
            s = MemHandleLock(h);
            StrNCopy(s,t,length+1);
            MemHandleUnlock(h);
            FldSetTextHandle(fld,h); // Restore the handle.
            FldDrawField(fld);
        }
        else
        {
            ErrFatalDisplay("Out of Memory!");
        }
    }
    else
    {
        ErrFatalDisplay("Null Field Pointer!");
    }
}

/*
 * Update a time field and redraw it.
 */
void
UpdateTimeField(FieldPtr fld, double num, int st)
{
    char        *s; 
    MemHandle   h;
    UInt16      size;
    double      hr, min;
    int         dayshift = 0;

    if (fld) {
        FldFreeMemory(fld);
        size = FldGetMaxChars(fld);
        h = MemHandleNew(size+1);

        if (h) {
            hr = floor(num);
            min = trunc((num - hr)*60.0);
    
            s = MemHandleLock(h);
            if (hr > 23.0) {
                dayshift = 1;
                hr = hr - 24.0;
            } else if (hr < 0.0) {
                dayshift = -1;
                hr = hr + 24.0;
            }
    
            TimeToAscii(hr,
                        min,
                        (st)?tfColon24h:sysPrefs.timeFormat,
                        s);
    
            if (dayshift > 0)
                StrCat(s,"++");
            else if (dayshift < 0)
                StrCat(s,"--");
    
            MemHandleUnlock(h);

            FldSetTextHandle(fld,h);
            FldDrawField(fld);
        } else {
            ErrFatalDisplay("Out of Memory!");
        }
    } else {
        ErrFatalDisplay("Null Field Pointer!");
    }
}

/*
 * Update a numeric field and redraw it.
 */
static void
UpdateNumField(FieldPtr fld, double num, int p)
{
    char        *s; 
    MemHandle   h;
    UInt16      size;

    if (fld) {
        FldFreeMemory(fld);
        size = FldGetMaxChars(fld);
        h = MemHandleNew(size+2);

        if (h) {
            s = MemHandleLock(h);
            StrDToA(s,num,p,size);
            MemHandleUnlock(h);
            FldSetTextHandle(fld,h);
            FldDrawField(fld);
        } else {
            ErrFatalDisplay("Out of Memory!");
        }
    } else {
        ErrFatalDisplay("Null Field Pointer!");
    }
}

/*
 * Update an editable field and redraw it.
 */
static void
UpdateEditField(FieldPtr fld, long num)
{
    char s[FIELDLENGTH];
    
    FldFreeMemory(fld); 
    StrIToA(s,num); 
    FldInsert(fld,s,StrLen(s)); 
    FldSetDirty(fld,true);
    FldDrawField(fld);
}

/*
 * Calculates the floating point lat/long from dms lat/long values
 */
static void
CalcOpts()
{
    if (moonPrefs.tz_h < 0)
        moonPrefs.tz = moonPrefs.tz_h - moonPrefs.tz_m / 60.0;
    else
        moonPrefs.tz = moonPrefs.tz_h + moonPrefs.tz_m / 60.0;

    if (moonPrefs.lat_d < 0)
        moonPrefs.lat = DegToRad(moonPrefs.lat_d - moonPrefs.lat_m / 60.0);
    else
        moonPrefs.lat = DegToRad(moonPrefs.lat_d + moonPrefs.lat_m / 60.0);

    if (moonPrefs.lng_d < 0)
        moonPrefs.lng = DegToRad(moonPrefs.lng_d - moonPrefs.lng_m / 60.0);
    else
        moonPrefs.lng = DegToRad(moonPrefs.lng_d + moonPrefs.lng_m / 60.0);
}

char datestr[FIELDLENGTH];
char timestr[FIELDLENGTH];

/*
 * Change the base date (used for various calculations) by 'days'.
 */
void
ChangeDate(ControlPtr dP, long days)
{
    DateType    dt;

    dt.year = moonPrefs.currentDT.year - 1904;
    dt.month = moonPrefs.currentDT.month;
    dt.day = moonPrefs.currentDT.day;
    
    DateAdjust(&dt, days);

    moonPrefs.currentDT.year = dt.year + 1904;
    moonPrefs.currentDT.month = dt.month;
    moonPrefs.currentDT.day = dt.day;
    
    DateToAscii(moonPrefs.currentDT.month,
                moonPrefs.currentDT.day,
                moonPrefs.currentDT.year,
                sysPrefs.dateFormat,
                datestr);

    if (dP) {
        CtlSetLabel(dP,
                    datestr);
        CtlDrawControl(dP);
    }
    
}

/*
 * Set the base date and TOD (used for various calculations) to 'now'.
 */
void
SetNow(ControlPtr dP, ControlPtr tP)
{
    UInt32      secs;
    
    /*
     * Set the date and time to now.
     */
    secs = TimGetSeconds();
    TimSecondsToDateTime(secs,
                         &moonPrefs.currentDT);

    /*
     * Round off to a 5 minute interval, because that's all
     * you can select with the time widget.
     */
    moonPrefs.currentDT.minute = (moonPrefs.currentDT.minute / 5) * 5;
    
    DateToAscii(moonPrefs.currentDT.month,
                moonPrefs.currentDT.day,
                moonPrefs.currentDT.year,
                sysPrefs.dateFormat,
                datestr);

    if (dP) {
        CtlSetLabel(dP,
                    datestr);
        CtlDrawControl(dP);
    }
    
    TimeToAscii(moonPrefs.currentDT.hour,
                moonPrefs.currentDT.minute,
                sysPrefs.timeFormat,
                timestr);
    if (tP) {
        CtlSetLabel(tP,
                    timestr);
        CtlDrawControl(tP);
    }
}

/*
 * Calculate the phase of the moon for the
 * current julian day and displays them.
 */
static void
ShowPhase(FormPtr formPtr)
{
    double        age, full, jday;
    double        gst, lst;
    double        UTr, UTs;
    double        ra, dec;
    double        alt, az;
    int           h,m,s;
    
    char          str1[FIELDLENGTH];

    BitmapPtr     bmp;

    FieldPtr      risePtr = (FieldPtr)FrmGetObjectPtr(formPtr,
                                                      FrmGetObjectIndex(formPtr,
                                                                        f_MRise));
    FieldPtr      setPtr = (FieldPtr)FrmGetObjectPtr(formPtr,
                                                     FrmGetObjectIndex(formPtr,
                                                                       f_MSet));
    
    jday = JDay(moonPrefs.currentDT.year,
                moonPrefs.currentDT.month,
                moonPrefs.currentDT.day,
                doubletime(moonPrefs.currentDT.hour,
                           moonPrefs.currentDT.minute,0),
                moonPrefs.tz);

    LunarPhase(&age, &full, jday);
    
    UpdateNumField((FieldPtr)FrmGetObjectPtr(formPtr,
                                             FrmGetObjectIndex(formPtr,
                                                               f_MAge)),
                   age*100,1);
    UpdateNumField((FieldPtr)FrmGetObjectPtr(formPtr,
                                             FrmGetObjectIndex(formPtr,
                                                               f_MFull)),
                   full*100,1);

    /*
     * Draw Bar graphs
     */
    bmp = MemHandleLock(bitmapHandle);
    WinDrawBitmap(bmp,108,49);
    WinDrawBitmap(bmp,108,62);
    MemHandleUnlock(bitmapHandle);
    
    bar.topLeft.y = 51;
    bar.extent.x = age * BARSIZE;
    WinDrawRectangle(&bar,0);
    
    bar.topLeft.y = 64;
    bar.extent.x =  full * BARSIZE;
    WinDrawRectangle(&bar,0);    
    
    LunarPos(&ra, &dec,
             jday, NULL);
    
    RadToDMS(ra/15.0,
             &h,&m,&s);
    StrPrintF(str1,"%dh %dm %ds",
              h,m,s);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_MRA)),
                    str1);
    
    RadToDMS(dec,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_MDec)),
                    str1);

    gst = JDayToGST(jday);
    lst = GSTToLST(gst,moonPrefs.lng);

    RaDecToAltAz(ra,dec,
                 lst,moonPrefs.lat,
                 &alt,&az);
    
    RadToDMS(alt,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_MAlt)),
                    str1);
    
    RadToDMS(az,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_MAz)),
                    str1);

    /*
     * MHEINZ: I moved this to the bottom of the function so
     * that all the fast-updating fields would get drawn before
     * this (relatively) slow calculation takes place.
     */
    if (MoonRiseSet(moonPrefs.currentDT.year,
                    moonPrefs.currentDT.month,
                    moonPrefs.currentDT.day,
                    moonPrefs.tz,
                    moonPrefs.lat,
                    moonPrefs.lng,
                    &UTr,
                    &UTs))
    {
        UpdateTextField(risePtr,"**:**");
        UpdateTextField(setPtr,"**:**");
    }
    else 
    {
        UpdateTimeField(risePtr,
                        UTToLT(UTr,moonPrefs.tz),0);
        UpdateTimeField(setPtr,
                        UTToLT(UTs,moonPrefs.tz),0);
    }
}

/*
 * Display the current Greenwich and Local Sidereal times
 */
static void
ShowLST(FormPtr formPtr)
{
    double jday, gst, lst;
    
    jday = JDay(moonPrefs.currentDT.year,
                moonPrefs.currentDT.month,
                moonPrefs.currentDT.day,
                doubletime(moonPrefs.currentDT.hour,
                           moonPrefs.currentDT.minute,0),
                moonPrefs.tz);

    gst = JDayToGST(jday);
    lst = GSTToLST(gst,moonPrefs.lng);
    
    UpdateNumField((FieldPtr)FrmGetObjectPtr(formPtr,
                                             FrmGetObjectIndex(formPtr,
                                                               f_Jday)),
                   jday,1);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_GST)),
                    gst,1);
    
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_LST)),
                    lst,1);
}

/*
 * Show the times for sunrise and sunset
 */
static void
ShowSolar(FormPtr formPtr)
{
    double      UTr, UTs;
    double      jday, gst, lst, l;
    double      ra, dec;
    double      alt, az;
    int         h,m,s;

    FieldPtr srisePtr = (FieldPtr)FrmGetObjectPtr(formPtr,
                                                  FrmGetObjectIndex(formPtr,
                                                                    f_SRise));
    FieldPtr ssetPtr = (FieldPtr)FrmGetObjectPtr(formPtr,
                                                 FrmGetObjectIndex(formPtr,
                                                                   f_SSet));
    FieldPtr sraPtr = (FieldPtr)FrmGetObjectPtr(formPtr,
                                                FrmGetObjectIndex(formPtr,
                                                                  f_SRA));
    FieldPtr sdcPtr = (FieldPtr)FrmGetObjectPtr(formPtr,
                                                FrmGetObjectIndex(formPtr,
                                                                  f_SDec));
    FieldPtr salPtr = (FieldPtr)FrmGetObjectPtr(formPtr,
                                                FrmGetObjectIndex(formPtr,
                                                                  f_SAlt));
    FieldPtr sazPtr = (FieldPtr)FrmGetObjectPtr(formPtr,
                                                FrmGetObjectIndex(formPtr,
                                                                  f_SAz));

    char str1[FIELDLENGTH];
    
    if (SunRiseSet(moonPrefs.currentDT.year,
                   moonPrefs.currentDT.month,
                   moonPrefs.currentDT.day,
                   moonPrefs.tz,
                   moonPrefs.lat,
                   moonPrefs.lng,
                   &UTr,
                   &UTs)) {
        UpdateTextField(srisePtr,"**:**");
        UpdateTextField(ssetPtr,"**:**");

        return;
    }

    UpdateTimeField(srisePtr,
                    UTToLT(UTr,moonPrefs.tz),0);
    UpdateTimeField(ssetPtr,
                    UTToLT(UTs,moonPrefs.tz),0);

    jday = JDay(moonPrefs.currentDT.year,
                moonPrefs.currentDT.month,
                moonPrefs.currentDT.day,
                doubletime(moonPrefs.currentDT.hour,
                           moonPrefs.currentDT.minute,0),
                moonPrefs.tz);
    
    gst = JDayToGST(jday);
    lst = GSTToLST(gst,moonPrefs.lng);

    /*
     * Longitude of the sun
     */
    l = SolarLong(jday,
                  NULL);

    /*
     * ra and dec for midnight, day 1
     */
    CelestialToRaDec(0.0,l,
                     jday,
                     &ra, &dec);
    
    RadToDMS(ra/15.0,
             &h,&m,&s);
    StrPrintF(str1,"%dh %dm %ds",
              h,m,s);
    UpdateTextField(sraPtr,str1);
    
    RadToDMS(dec,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField(sdcPtr,str1);

    RaDecToAltAz(ra,dec,
                 lst,moonPrefs.lat,
                 &alt,&az);
    
    RadToDMS(alt,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField(salPtr,str1);
    
    RadToDMS(az,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField(sazPtr,str1);
}

/*
 * Process all menu selections. This is possible because
 * all Astro forms use the same menu.
 */
Boolean
DoMenu(UInt16 itemID)
{
    Boolean handled = false;
    
    switch (itemID) {

        case mi_About:
            FrmAlert(a_About);
            handled = true;
            break;

        case mi_Help:
            switch (moonPrefs.activeFormId) 
            {
                case f_Moon:
                    FrmAlert(a_MoonHelp);
                    break;
                case f_LST:
                    FrmAlert(a_LSTHelp);
                    break;
                case f_Solar:
                    FrmAlert(a_SolarHelp);
                    break;
                case f_Planet:
                    FrmAlert(a_PlanetHelp);
                    break;
                case f_Catalog:
                    FrmAlert(a_CatalogHelp);
                    break;
                default:
                    ErrFatalDisplay("Unknown form. Missing Help!");
            }
            handled = true;
            break;
        case mi_Options:
            if (moonPrefs.activeFormId != f_Options) 
            {
                saveFormId = moonPrefs.activeFormId;
                FrmPopupForm(f_Options);
            }
            handled = true;
            break;

        case mi_Catalog:
            if (moonPrefs.activeFormId != f_Catalog) 
            {
                saveFormId = moonPrefs.activeFormId;
                FrmGotoForm(f_Catalog);
            }
            handled = true;
            break;
                
        case mi_Phase:
            if (moonPrefs.activeFormId != f_Moon)
                FrmGotoForm(f_Moon);
            handled = true;
            break;

        case mi_LST:
            if (moonPrefs.activeFormId != f_LST)
                FrmGotoForm(f_LST);
            handled = true;
            break;

        case mi_Solar:
            if (moonPrefs.activeFormId != f_Solar)
                FrmGotoForm(f_Solar);
            handled = true;
            break;

        case mi_Planet:
            if (moonPrefs.activeFormId != f_Planet)
                FrmGotoForm(f_Planet);
            handled = true;
            break;
            
        case mi_Night:
            if (colorSupport)
            {
                // Color support is enabled. Toggle night mode.
                moonPrefs.nightMode = !moonPrefs.nightMode;
                NightMode();
            }
    }

    return handled;
}

/*
 * Display the date selector and update the current date with the
 * user's selection. dP is a pointer to the selector resource
 * to refresh.
 */
void
SetDate(ControlPtr dP)
{    
    if (SelectDay(selectDayByDay,
                  &moonPrefs.currentDT.month,
                  &moonPrefs.currentDT.day,
                  &moonPrefs.currentDT.year,
                  "Select Date")) {
        DateToAscii(moonPrefs.currentDT.month,
                    moonPrefs.currentDT.day,
                    moonPrefs.currentDT.year,
                    sysPrefs.dateFormat,
                    datestr);
        CtlSetLabel(dP,
                    datestr);
        CtlDrawControl(dP);
                
    }
}

/*
 * Display the time selector and update the time with the user's
 * selection.
 *
 * tP is a pointer to the selector resource to refresh.
 */
void
SetTime(ControlPtr tP)
{
    TimeType    s,e;
    e.hours = s.hours = moonPrefs.currentDT.hour;
    e.minutes = s.minutes = moonPrefs.currentDT.minute;
    if (SelectTimeV33(&s,&e,
                      false,
                      "Set Start Time Only",
                      0) && *((Int16*)&s)!=-1) {
        moonPrefs.currentDT.hour = s.hours;
        moonPrefs.currentDT.minute = s.minutes;
                
        TimeToAscii(moonPrefs.currentDT.hour,
                    moonPrefs.currentDT.minute,
                    sysPrefs.timeFormat,
                    timestr);
        CtlSetLabel(tP,
                    timestr);
        CtlDrawControl(tP);
    }
}

char *PlanetName[] = {"Mercury",
                      "Venus",
                      "Earth",
                      "Mars",
                      "Jupiter",
                      "Saturn",
                      "Uranus",
                      "Neptune",
                      "Pluto"};

/*
 * Display the co-ordinates for a planet
 */
static int planet_num;

static void
ShowPlanetPos(FormPtr formPtr)
{
    double ra, dec, jday;
    double gst, lst;
    double alt, az;
    
    int h,m,s;
    
    char str1[FIELDLENGTH];

    jday = JDay(moonPrefs.currentDT.year,
                moonPrefs.currentDT.month,
                moonPrefs.currentDT.day,
                doubletime(moonPrefs.currentDT.hour,
                           moonPrefs.currentDT.minute,0),
                moonPrefs.tz);

    gst = JDayToGST(jday);
    lst = GSTToLST(gst,moonPrefs.lng);

    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                               FrmGetObjectIndex(formPtr,
                                                                 f_PosPlanet)),
                    PlanetName[planet_num]);
    
    PlanetPos(&ra,&dec,
              planet_num,
              jday);

    RadToDMS(ra/15.0,
             &h,&m,&s);
    StrPrintF(str1,"%dh %dm %ds",
              h,m,s);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                       FrmGetObjectIndex(formPtr,
                                                         f_PosRA))
                    ,str1);
    
    RadToDMS(dec,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                               FrmGetObjectIndex(formPtr,
                                                                 f_PosDec)),
                    str1);
    
    RaDecToAltAz(ra,dec,
                 lst,moonPrefs.lat,
                 &alt,&az);
    
    RadToDMS(alt,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                               FrmGetObjectIndex(formPtr,
                                                                 f_PosAlt)),
                    str1);
    
    RadToDMS(az,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                               FrmGetObjectIndex(formPtr,
                                                                 f_PosAz)),
                    str1);
}

/*
 * Display the rising and setting times for the 8 planets.
 */
static void
ShowPlanet(FormPtr formPtr)
{
    double      UTr, UTs;
    
    PlanetRiseSet(MERCURY,
                  moonPrefs.currentDT.year,
                  moonPrefs.currentDT.month,
                  moonPrefs.currentDT.day,
                  moonPrefs.lat,
                  moonPrefs.lng,
                  &UTr, &UTs);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_MRise)),
                    UTToLT(UTr,moonPrefs.tz),0);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_MSet)),
                    UTToLT(UTs,moonPrefs.tz),0);

    PlanetRiseSet(VENUS,
                  moonPrefs.currentDT.year,
                  moonPrefs.currentDT.month,
                  moonPrefs.currentDT.day,
                  moonPrefs.lat,
                  moonPrefs.lng,
                  &UTr, &UTs);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_VRise)),
                    UTToLT(UTr,moonPrefs.tz),0);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_VSet)),
                    UTToLT(UTs,moonPrefs.tz),0);
    
    PlanetRiseSet(MARS, 
                  moonPrefs.currentDT.year,
                  moonPrefs.currentDT.month,
                  moonPrefs.currentDT.day,
                  moonPrefs.lat,
                  moonPrefs.lng,
                  &UTr, &UTs);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_ARise)),
                    UTToLT(UTr,moonPrefs.tz),0);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_ASet)),
                        UTToLT(UTs,moonPrefs.tz),0);
    
    PlanetRiseSet(JUPITER,
                  moonPrefs.currentDT.year,
                  moonPrefs.currentDT.month,
                  moonPrefs.currentDT.day,
                  moonPrefs.lat,
                  moonPrefs.lng,
                  &UTr, &UTs);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_JRise)),
                    UTToLT(UTr,moonPrefs.tz),0);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_JSet)),
                    UTToLT(UTs,moonPrefs.tz),0);
    
    PlanetRiseSet(SATURN,
                  moonPrefs.currentDT.year,
                  moonPrefs.currentDT.month,
                  moonPrefs.currentDT.day,
                  moonPrefs.lat,
                  moonPrefs.lng,
                  &UTr, &UTs);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_TRise)),
                    UTToLT(UTr,moonPrefs.tz),0);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_TSet)),
                    UTToLT(UTs,moonPrefs.tz),0);
    
    PlanetRiseSet(URANUS,
                  moonPrefs.currentDT.year,
                  moonPrefs.currentDT.month,
                  moonPrefs.currentDT.day,
                  moonPrefs.lat,
                  moonPrefs.lng,
                  &UTr, &UTs);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_URise)),
                    UTToLT(UTr,moonPrefs.tz),0);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                                 FrmGetObjectIndex(formPtr,
                                                                   f_USet)),
                    UTToLT(UTs,moonPrefs.tz),0);
    
    PlanetRiseSet(NEPTUNE,
                  moonPrefs.currentDT.year,
                  moonPrefs.currentDT.month,
                  moonPrefs.currentDT.day,
                  moonPrefs.lat,
                  moonPrefs.lng,
                  &UTr, &UTs);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_NRise)),
                    UTToLT(UTr,moonPrefs.tz),0);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_NSet)),
                    UTToLT(UTs,moonPrefs.tz),0);

    PlanetRiseSet(PLUTO, 
                  moonPrefs.currentDT.year,
                  moonPrefs.currentDT.month,
                  moonPrefs.currentDT.day,
                  moonPrefs.lat,
                  moonPrefs.lng,
                  &UTr, &UTs);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_PRise)),
                    UTToLT(UTr,moonPrefs.tz),0);
    UpdateTimeField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_PSet)),
                    UTToLT(UTs,moonPrefs.tz),0);
}

/*
 * Events for the planetary rise/set form
 */
static Boolean
PlanetFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();

    switch (event->eType) {
        case ctlSelectEvent:
            /*
             * Process various button clicks. If the name of a planet
             * is clicked, open the position dialog box for that
             * planet.
             */
            saveFormId = moonPrefs.activeFormId;
            switch (event->data.ctlSelect.controlID) {
                case s_Date4:
                    SetDate(date4Ptr);
                    ShowPlanet(formPtr);
                    break;
                case s_Time4:
                    SetTime(time4Ptr);
                    ShowPlanet(formPtr);
                    break;
                case b_Now:
                    SetNow(date4Ptr,time4Ptr);
                    ShowPlanet(formPtr);
                    break;
                case b_Mercury:
                    saveFormId = moonPrefs.activeFormId;
                    planet_num=MERCURY; FrmPopupForm(f_PlanetPosition);
                    break;
                case b_Venus:
                    planet_num=VENUS; FrmPopupForm(f_PlanetPosition);
                    break;
                case b_Mars:
                    planet_num=MARS; FrmPopupForm(f_PlanetPosition);
                    break;
                case b_Jupiter:
                    planet_num=JUPITER; FrmPopupForm(f_PlanetPosition);
                    break;
                case b_Saturn:
                    planet_num=SATURN; FrmPopupForm(f_PlanetPosition);
                    break;
                case b_Uranus:
                    planet_num=URANUS; FrmPopupForm(f_PlanetPosition);
                    break;
                case b_Neptune:
                    planet_num=NEPTUNE; FrmPopupForm(f_PlanetPosition);
                    break;
                case b_Pluto:
                    planet_num=PLUTO; FrmPopupForm(f_PlanetPosition);
                    break;
            }
            handled = true;
            break;

        case ctlRepeatEvent:
            if (event->data.ctlSelect.controlID == r_Dated) {
                ChangeDate(date4Ptr,-1);
            } else if (event->data.ctlSelect.controlID == r_Dateu) {
                ChangeDate(date4Ptr,1);
            }
            ShowPlanet(formPtr);
            break;
        
        case frmOpenEvent:
            FrmSetMenu(formPtr,mainMenu);
            
            FrmDrawForm(formPtr);

            date4Ptr = (ControlPtr)FrmGetObjectPtr(formPtr,
                                                   FrmGetObjectIndex(formPtr,
                                                                     s_Date4));

            time4Ptr = (ControlPtr)FrmGetObjectPtr(formPtr,
                                                   FrmGetObjectIndex(formPtr,
                                                                     s_Time4));

            CtlSetLabel(date4Ptr,
                        datestr);
            CtlSetLabel(time4Ptr,
                        timestr);

            CtlDrawControl(date4Ptr);
            CtlDrawControl(time4Ptr);

            ShowPlanet(formPtr);

            handled = true;
            break;

        case menuEvent:
            handled = DoMenu(event->data.menu.itemID);
            break;

        default:
            handled = false;
    }

    return handled;
}

/*
 * Events for the sunrise form
 */
static Boolean
SolarFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();
    
    switch (event->eType) {
        case ctlSelectEvent:
            /* Process various button clicks. */
            switch (event->data.ctlSelect.controlID) {
                case s_Date3:
                    SetDate(date3Ptr);
                    break;
                case s_Time3:
                    SetTime(time3Ptr);
                    break;
                case b_Now:
                    SetNow(date3Ptr, time3Ptr);            
                    break;
            }
            ShowSolar(formPtr);
            handled = true;
            break;

        case ctlRepeatEvent:
            if (event->data.ctlSelect.controlID == r_Dated) {
                ChangeDate(date3Ptr,-1);
            } else if (event->data.ctlSelect.controlID == r_Dateu) {
                ChangeDate(date3Ptr,1);
            }
            ShowSolar(formPtr);
            break;

        case frmOpenEvent:
            FrmSetMenu(formPtr,mainMenu);            
        
            FrmDrawForm(formPtr);

            date3Ptr = (ControlPtr)FrmGetObjectPtr(formPtr,
                                                   FrmGetObjectIndex(formPtr,
                                                                     s_Date3));

            time3Ptr = (ControlPtr)FrmGetObjectPtr(formPtr,
                                                   FrmGetObjectIndex(formPtr,
                                                                     s_Time3));

            CtlSetLabel(date3Ptr,
                        datestr);
            CtlSetLabel(time3Ptr,
                        timestr);

            CtlDrawControl(date3Ptr);
            CtlDrawControl(time3Ptr);

            ShowSolar(formPtr);

            handled = true;
            break;

        case menuEvent:
            handled = DoMenu(event->data.menu.itemID);
            break;

        default:
            handled = false;
    }

    return handled;
}

/*
 * Events for the sidereal time form
 */
Boolean
LSTFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();
    
    switch (event->eType) {
        case ctlSelectEvent:
            /* Process various button clicks. */
            switch (event->data.ctlSelect.controlID) {
                case s_Date2:
                    SetDate(date2Ptr);
                    break;
                case s_Time2:
                    SetTime(time2Ptr);
                    break;
                case b_Now:
                    SetNow(date2Ptr, time2Ptr);            
                    break;
            }
            ShowLST(formPtr);
            handled = true;
            break;

        case ctlRepeatEvent:
            if (event->data.ctlSelect.controlID == r_Dated) {
                ChangeDate(date2Ptr,-1);
            } else if (event->data.ctlSelect.controlID == r_Dateu) {
                ChangeDate(date2Ptr,1);
            }
            ShowLST(formPtr);
            break;

        case frmOpenEvent:
            FrmSetMenu(formPtr,mainMenu);            
        
            FrmDrawForm(formPtr);

            date2Ptr = (ControlPtr)FrmGetObjectPtr(formPtr,
                                                   FrmGetObjectIndex(formPtr,
                                                                     s_Date2));

            time2Ptr = (ControlPtr)FrmGetObjectPtr(formPtr,
                                                   FrmGetObjectIndex(formPtr,
                                                                     s_Time2));

            CtlSetLabel(date2Ptr,
                        datestr);
            CtlSetLabel(time2Ptr,
                        timestr);

            CtlDrawControl(date2Ptr);
            CtlDrawControl(time2Ptr);

            ShowLST(formPtr);

            handled = true;
            break;

        case menuEvent:
            handled = DoMenu(event->data.menu.itemID);
            break;

        default:
            handled = false;
    }

    return handled;
}

/*
 * Events for the lunar phase form
 */
Boolean
MoonFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();
    
    switch (event->eType) {
        case ctlSelectEvent:
            /* Process various button clicks. */
            switch (event->data.ctlSelect.controlID) {
                case s_Date:
                    SetDate(datePtr);
                    break;
                case s_Time:
                    SetTime(timePtr);
                    break;
                case b_Now:
                    SetNow(datePtr, timePtr);            
                    break;
            }
            ShowPhase(formPtr);
            handled = true;
            break;

        case ctlRepeatEvent:
            if (event->data.ctlSelect.controlID == r_Dated) {
                ChangeDate(datePtr,-1);
            } else if (event->data.ctlSelect.controlID == r_Dateu) {
                ChangeDate(datePtr,1);
            }
            ShowPhase(formPtr);
            break;

        case frmOpenEvent:
            FrmSetMenu(formPtr,mainMenu);            
        
            FrmDrawForm(formPtr);

            datePtr = (ControlPtr)FrmGetObjectPtr(formPtr,
                                                  FrmGetObjectIndex(formPtr,
                                                                    s_Date));
            timePtr = (ControlPtr)FrmGetObjectPtr(formPtr,
                                                  FrmGetObjectIndex(formPtr,
                                                                    s_Time));
            CtlSetLabel(datePtr,
                        datestr);
            CtlDrawControl(datePtr);

            CtlSetLabel(timePtr,
                        timestr);
            CtlDrawControl(timePtr);
        
            ShowPhase(formPtr);

            handled = true;
            break;

        case menuEvent:
            handled = DoMenu(event->data.menu.itemID);
            break;
        
        default:
            handled = false;
    }

    return handled;
}

/*
 * Events for the options form
 */
static Boolean
OptionsFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();
    
    switch (event->eType) {
        case ctlSelectEvent:
            if (event->data.ctlSelect.controlID == b_OK) {
                moonPrefs.tz_h = GetNumField(formPtr,f_TZh);
                moonPrefs.tz_m = GetNumField(formPtr,f_TZm);
                if (CtlGetValue((ControlPtr)FrmGetObjectPtr(formPtr,
                                                            FrmGetObjectIndex(formPtr,
                                                                              c_TZw))))
                    moonPrefs.tz_h = -moonPrefs.tz_h;
        
                moonPrefs.lat_d = GetNumField(formPtr,f_LATd);
                moonPrefs.lat_m = GetNumField(formPtr,f_LATm);
                if (CtlGetValue((ControlPtr)FrmGetObjectPtr(formPtr,
                                                            FrmGetObjectIndex(formPtr,
                                                                              c_LATs))))
                    moonPrefs.lat_d = -moonPrefs.lat_d;
        
                moonPrefs.lng_d = GetNumField(formPtr,f_LNGd);
                moonPrefs.lng_m = GetNumField(formPtr,f_LNGm);
                if (!CtlGetValue((ControlPtr)FrmGetObjectPtr(formPtr,
                                                             FrmGetObjectIndex(formPtr,
                                                                               c_LNGw))))
                    moonPrefs.lng_d = -moonPrefs.lng_d;

                CalcOpts();

                /*
                 * Restore the previous form.
                 */
                moonPrefs.activeFormId = saveFormId;
                FrmReturnToForm(saveFormId);
                handled = true;
            }
            
            break;
         
        case frmOpenEvent:

            FrmDrawForm(formPtr);

            UpdateEditField((FieldPtr)FrmGetObjectPtr(formPtr,
                                                      FrmGetObjectIndex(formPtr,
                                                                        f_TZh)),
                            abs(moonPrefs.tz_h));
            UpdateEditField((FieldPtr)FrmGetObjectPtr(formPtr,
                                                      FrmGetObjectIndex(formPtr,
                                                                        f_TZm)),
                            moonPrefs.tz_m);
            CtlSetValue((ControlPtr)FrmGetObjectPtr(formPtr,
                                                    FrmGetObjectIndex(formPtr,
                                                                      c_TZw)),
                        (moonPrefs.tz < 0));

            UpdateEditField((FieldPtr)FrmGetObjectPtr(formPtr,
                                                      FrmGetObjectIndex(formPtr,
                                                                        f_LATd)),
                            abs(moonPrefs.lat_d));
            UpdateEditField((FieldPtr)FrmGetObjectPtr(formPtr,
                                                      FrmGetObjectIndex(formPtr,
                                                                        f_LATm)),
                            moonPrefs.lat_m);
            CtlSetValue((ControlPtr)FrmGetObjectPtr(formPtr,
                                                    FrmGetObjectIndex(formPtr,
                                                                      c_LATs)),
                        (moonPrefs.lat < 0));

            UpdateEditField((FieldPtr)FrmGetObjectPtr(formPtr,
                                                      FrmGetObjectIndex(formPtr,
                                                                        f_LNGd)),
                            abs(moonPrefs.lng_d));
            UpdateEditField((FieldPtr)FrmGetObjectPtr(formPtr,
                                                      FrmGetObjectIndex(formPtr,
                                                                        f_LNGm)),
                            moonPrefs.lng_m);
            CtlSetValue((ControlPtr)FrmGetObjectPtr(formPtr,
                                                    FrmGetObjectIndex(formPtr,
                                                                      c_LNGw)),
                        (moonPrefs.lng > 0));
        
            handled = true;
            break;

        default:
            handled = false;
    }

    return handled;
}

/*
 * Events for the Position form
 */
static Boolean
PositionFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();
    
    switch (event->eType) {
        case ctlSelectEvent:
            if (event->data.ctlSelect.controlID == b_OK) {
                moonPrefs.activeFormId = saveFormId;
                FrmReturnToForm(saveFormId);
                handled = true;
            }
            break;
         
        case frmOpenEvent:
        
            FrmDrawForm(formPtr);
            ShowPlanetPos(formPtr);
            handled = true;
            break;

        default:
            handled = false;
    }

    return handled;
}

/*
 * Events for the main application (basically, setting form event
 * handlers.)
 */
static Boolean
ApplicationHandleEvent(EventPtr event)
{
    FormPtr form;
    Boolean handled = false;

    if (event->eType == frmLoadEvent) {
        moonPrefs.activeFormId = event->data.frmLoad.formID;
        form = FrmInitForm(moonPrefs.activeFormId);
        FrmSetActiveForm(form);
        
        switch (moonPrefs.activeFormId) {
            case f_Moon:
                FrmSetEventHandler(form,
                                   MoonFormHandleEvent);
                handled = true;
                break;

            case f_Options:
                FrmSetEventHandler(form,
                                   OptionsFormHandleEvent);
                handled = true;
                break;

            case f_Catalog:
                FrmSetEventHandler(form,
                                   CatalogFormHandleEvent);
                handled = true;
                break;
                
            case f_SelectCatalog:
                FrmSetEventHandler(form,
                                   SelectCatalogFormHandleEvent);
                handled = true;
                break;
                
            case f_FindObject:
                FrmSetEventHandler(form,
                                   FindObjectFormHandleEvent);
                handled = true;
                break;
                
            case f_ObjectPosition:
                FrmSetEventHandler(form,
                                   ObjectPositionFormHandleEvent);
                handled = true;
                break;
                
            case f_PlanetPosition:
                FrmSetEventHandler(form,
                                   PositionFormHandleEvent);
                handled = true;
                break;
            
            case f_LST:
                FrmSetEventHandler(form,
                                   LSTFormHandleEvent);
                handled = true;
                break;
            
            case f_Solar:
                FrmSetEventHandler(form,
                                   SolarFormHandleEvent);
                handled = true;
                break;

            case f_Planet:
                FrmSetEventHandler(form,
                                   PlanetFormHandleEvent);
                handled = true;
                break;
        }
    }
    return handled;
}

/*
 * Application event loop. duh.
 */
static void
EventLoop(void)
{
    EventType   event;
    UInt16      error;

    do {
        EvtGetEvent(&event,
                    evtWaitForever);
        if (!SysHandleEvent(&event))
            if (!MenuHandleEvent(NULL,
                                 &event,
                                 &error))
                if (!ApplicationHandleEvent(&event))
                    FrmDispatchEvent(&event);
    } while (event.eType != appStopEvent);
}

/*
 * Initialize the math library; allocate the handles for
 * various text fields; load our previously saved preferences.
 */
static void
StartApplication(void)
{
    PrefGetPreferences(&sysPrefs);

    if ((PrefGetAppPreferences(APPID,
                              0,
                              &moonPrefs,
                              &mptSize,
                              true) == noPreferenceFound) ||
        mptSize != sizeof(moonPrefs) ||
        moonPrefs.version!= ASTROVERSION) {
        /*
         * New install or else preferences are old.  Use default
         * settings
         */
        moonPrefs.version = ASTROVERSION;
        moonPrefs.tz = -4.0; moonPrefs.tz_h = -4; moonPrefs.tz_m = 0;
        moonPrefs.lat = 41.25; moonPrefs.lat_d = 41; moonPrefs.lat_m = 15;
        moonPrefs.lng = 75.5; moonPrefs.lng_d = 75; moonPrefs.lng_m = 30;
        moonPrefs.activeFormId = f_Moon;
        moonPrefs.nightMode = false;
        moonPrefs.catCardNo = 0;
        moonPrefs.catDbName[0] = 0;
        
        SetNow(NULL,NULL); /* Set saved time */
        CalcOpts(); /* Set floating point lat/long */
    } else {
        TimeToAscii(moonPrefs.currentDT.hour,
                    moonPrefs.currentDT.minute,
                    sysPrefs.timeFormat,
                    timestr);
            
        DateToAscii(moonPrefs.currentDT.month,
                    moonPrefs.currentDT.day,
                    moonPrefs.currentDT.year,
                    sysPrefs.dateFormat,
                    datestr);
    }

    /*
     * Handle for moon form bitmap
     */
    bitmapHandle = DmGetResource('Tbmp',bmp_Bar);
    ErrFatalDisplayIf(!bitmapHandle,
                      "Could not find the bitmap!");
    /*
     * Sanity check.  If the app was terminated
     * from inside a dialog box, reset to default
     * form.
     */
    if (moonPrefs.activeFormId == f_Options)
        moonPrefs.activeFormId = f_Moon;
    else if (moonPrefs.activeFormId == f_PlanetPosition)
        moonPrefs.activeFormId = f_Planet;
    else if (moonPrefs.activeFormId == f_ObjectPosition ||
        moonPrefs.activeFormId == f_FindObject ||
        moonPrefs.activeFormId == f_SelectCatalog)
        moonPrefs.activeFormId = f_Catalog;

    if (CheckMinRomVersion(sysVersion35)) 
    {
        WinScreenMode(winScreenModeGetSupportsColor,
                      NULL, NULL, NULL,
                      &colorSupport);
        if (colorSupport) 
        {
            UInt32 depth;
            WinScreenMode(winScreenModeGet,
                          NULL, NULL, &depth, NULL);
            colorSupport = (depth >= 8);
            if (colorSupport)
                mainMenu=m_Color;
        }
    }

    saveFormId = moonPrefs.activeFormId;
    FrmGotoForm(moonPrefs.activeFormId);
}

/*
 * Releases all memory; closes libraries; saves preferences
 */
static void
StopApplication()
{
    Err         error;
    UInt16      usecount;

    FreeCat();
    
    FrmCloseAllForms();

    PrefSetAppPreferences(APPID,
                          0, 1,
                          &moonPrefs,
                          sizeof(moonPrefs),
                          true);
    
    DmReleaseResource(bitmapHandle);
    
    error = MathLibClose(MathLibRef, &usecount);
    ErrFatalDisplayIf(error, "Can't close MathLib");
    if (usecount == 0)
        SysLibRemove(MathLibRef);

}

UInt32
PilotMain(UInt16 cmd,
          MemPtr cmdBPD,
          UInt16 launchFlags)
{
    Err         error;
    
    if (cmd == sysAppLaunchCmdNormalLaunch)
    {        
        error = SysLibFind(MathLibName, &MathLibRef);
        if (error)
            error = SysLibLoad(LibType, MathLibCreator, &MathLibRef);
        
        ErrFatalDisplayIf(error,
                          "MATHLIB NOT FOUND. You must install MathLib.prc To use this program.");
        
        error = MathLibOpen(MathLibRef, MathLibVersion);
        ErrFatalDisplayIf(error,
                          "MATHLIB NOT LOADED. MathLib.prc could not be opened.");
        
        // Main application
        StartApplication();
        EventLoop();
        StopApplication();
        // End of main application.
        
        return 0;
    }

    return 0;
}

/*
 * This color table defines a set things to be drawn in red.
 */
static UIColorTableEntries redList[] = {
    UIObjectFrame, UIObjectForeground, UIObjectSelectedFill,
    UIMenuFrame, UIMenuForeground, UIMenuSelectedFill,
    UIFieldText, UIFieldCaret, UIFieldTextHighlightBackground,
    UIFieldFepRawText, UIFieldFepConvertedText,
    UIFieldFepUnderline, UIFormFrame,
    UIDialogFrame, UIAlertFrame, UIOK, UICaution, UIWarning,
    UILastColorTableEntry 
};

/*
 * My version of a color table.
 */
typedef struct MyColorTableType
{
    UInt16 numEntries;
    RGBColorType rgb[UILastColorTableEntry];
    UInt32 appid;
} MyColorTableType;

/*
 * Toggles the red-on-black color scheme for color palm pilots.
 *
 * Follows the following logic:
 * 1) if no color table exists, create a night vision table.
 * 2) if a table exists, and it is *ours* delete it.
 *    if a saved table exists, restore it.
 * 3) if a table exists, and it is *not ours* save it.
 *    create a night vision table.
 *
 * Color tables are saved by renaming them from type sysResTSysPref
 * to type MWH2.
 */
void
NightMode()
{
    DmOpenRef savedPrefs = NULL;
    MemHandle handle = NULL;
    MyColorTableType *table;
    Boolean deleteCurrent = false;
    Boolean makeNew = false;
    
    Err err;
    UInt16 length = UILastColorTableEntry;
    UInt32 appId = APPID;
    
    int i;    
    
    RGBColorType redColor = { 0, 0xA0, 0, 0 };
    RGBColorType blackColor = { 0, 0, 0, 0 };
    
    Int16 indexCurrent, indexSave;

    savedPrefs=DmOpenDatabaseByTypeCreator(sysFileTSavedPreferences,
                                           sysFileCSystem,
                                           dmModeReadWrite);
    ErrFatalDisplayIf(savedPrefs==0,"Couldn't open Saved Preferences!");
    
    indexCurrent = DmFindResource(savedPrefs,
                                  sysResTSysPref,
                                  sysResIDPrefUIColorTable8,
                                  NULL);
    indexSave = DmFindResource(savedPrefs,
                               APPID,
                               sysResIDPrefUIColorTable8,
                               NULL);
    if (indexCurrent<0)
    {
        // No current table. Make a new one.
        makeNew = true;
        if (indexSave>=0) 
        {
            // Delete the old save table.
            err = DmRemoveResource(savedPrefs,indexSave);
            ErrFatalDisplayIf(err,"Failed to remove color table!");
        }    
    }
    else
    {
        // A table exists. Is it one of ours?
        handle = DmGetResourceIndex(savedPrefs,indexCurrent);
        if (MemHandleSize(handle)==sizeof(struct MyColorTableType)) 
        {
            table = MemHandleLock(handle);
            deleteCurrent = (table->appid == APPID);
            MemHandleUnlock(handle);
        }
        
        // Turn night mode off.
        if (deleteCurrent) 
        {
            // Delete the red-on-black color table.
            err = DmRemoveResource(savedPrefs,indexCurrent);
            ErrFatalDisplayIf(err,"Failed to remove color table!");

            if (indexSave>=0) 
            {
                // Restore saved table.
                DmResType rt = sysResTSysPref;
            
                err = DmSetResourceInfo(savedPrefs,
                                        indexSave,
                                        &rt,
                                        NULL);
            }
        }
        else 
        {
            //Save the current table.

            if (indexSave>=0) 
            {
                // Delete the old save table.
                err = DmRemoveResource(savedPrefs,indexSave);
                ErrFatalDisplayIf(err,"Failed to remove color table!");
            }
            
            // Save the current table, by renaming it.
            err = DmSetResourceInfo(savedPrefs,
                                    indexCurrent,
                                    &appId,
                                    NULL);
            ErrFatalDisplayIf(err,"Failed to save color table!");
            
            makeNew = true;
        }
    }

    if (makeNew)
    {
        /*
         * Turn night mode on.  Set up the new color scheme!
         */
        handle = DmNewResource(savedPrefs,
                               sysResTSysPref,
                               sysResIDPrefUIColorTable8,
                               sizeof(MyColorTableType));
        ErrFatalDisplayIf(handle==NULL,"Couldn't create a Preference!");
        
        table = MemHandleLock(handle);

        /*
         * Set the color table length field
         */
        err=DmWrite(table,
                    OffsetOf(MyColorTableType,numEntries),
                    &length,
                    sizeof(UInt16));

        ErrFatalDisplayIf(err,"DMWrite Failed!");

        /*
         * Set our special "cookie" so that we can identify 
         * color tables that we created.
         */
        err=DmWrite(table,
                    OffsetOf(MyColorTableType,appid),
                    &appId,
                    sizeof(UInt32));

        /*
         * Set the entire table to black.
         */
        for (i=0; i<UILastColorTableEntry; i++)
            DmWrite(table,
                    OffsetOf(MyColorTableType,rgb[i]),
                    &blackColor,sizeof(RGBColorType));
        /*
         * Set all entries in the redlist to red!
         */
        for (i=0; redList[i]!=UILastColorTableEntry; i++)
            DmWrite(table,
                    OffsetOf(MyColorTableType,rgb[redList[i]]),
                    &redColor,sizeof(RGBColorType));

        MemHandleUnlock(handle);
    
        DmReleaseResource(handle);
    }

    DmCloseDatabase(savedPrefs);

    GotoApplication(APPID);
}

