/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: AstroCat.c,v 1.6 2000/09/27 20:36:56 mheinz Exp $
 * Copyright (C) 2000, Michael Heinz
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include <PalmOS.h>
#include "PalmUtil.h"
#include "resource.h"
#include "AstroCat.h"
#include "AstroLib.h"
#include "Astro.h"

extern void * memcpy(void *dst,const void *src,unsigned long l);

/*
 * Information about currently selected catalog
 */
static AstroAppInfoType* aaInfo = NULL;
static MemHandle       catList = NULL;
static MemHandle       catBuf = NULL;
static DmOpenRef       catDB = NULL;
static LocalID         catDbID = NULL;
static UInt16          catCurrRecord = 0;

/*
 * List of installed catalogs
 */
static DBRecord**      recordList = NULL;
static int             recordCount = 0;

/*
 * Pointers to catalog objects
 */
static ListPtr         catNamePtr;
static ListPtr         catalogPtr;
static ControlPtr      date5Ptr, time5Ptr;

/*
 * Events for the Object Position form
 */
Boolean
ObjectPositionFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();
    
    switch (event->eType) {
        case ctlSelectEvent:
            if (event->data.ctlSelect.controlID == b_OK)
            {
                moonPrefs.activeFormId = saveFormId;
                FrmReturnToForm(saveFormId);
                handled = true;
            }
            break;

        case frmOpenEvent:
        
            FrmDrawForm(formPtr);
            LoadObject(formPtr);
            break;

        default:
            handled = false;
    }

    return handled;
}

// Global used to hold last find text.
static char findString[FIELDLENGTH] = { "\0" };

/*
 * Events for the FindObject form
 */
Boolean
FindObjectFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();
    
    switch (event->eType) {
        case ctlSelectEvent:
            if (event->data.ctlSelect.controlID == b_OK) {
                char *newString = FldGetTextPtr((FieldPtr)FrmGetObjectPtr(formPtr,FrmGetObjectIndex(formPtr,f_Find)));
                if (newString != NULL) 
                {
                    StrToLower(findString,newString);
                    if (catDB != NULL)
                        FindObject(0);
                }                
                else 
                {
                    findString[0]='\0';
                }
                
                moonPrefs.activeFormId = saveFormId;
                FrmReturnToForm(saveFormId);
                handled = true;
            }
            else if (event->data.ctlSelect.controlID == b_Cancel)
            {
                moonPrefs.activeFormId = saveFormId;
                FrmReturnToForm(saveFormId);
                handled = true;
            }

            break;

        case frmOpenEvent:
        
            FrmDrawForm(formPtr);            
            handled = true;
            break;

        default:
            handled = false;
    }

    return handled;
}

/*
 * Events for the SelectCatalog form
 */
Boolean
SelectCatalogFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();
    
    switch (event->eType) {
        case ctlSelectEvent:
            if (event->data.ctlSelect.controlID == b_OK) {
                int i = LstGetSelection(catNamePtr);
                if (i != noListSelection) 
                {
                    moonPrefs.catCardNo = recordList[i]->cardNo;
                    catDbID = recordList[i]->dbID;
                    StrCopy(moonPrefs.catDbName,recordList[i]->name);
                    LoadCatalog();
                    moonPrefs.activeFormId = saveFormId;
                    if (recordList != NULL)
                        FreeDBList(recordList,recordCount);
                    FrmReturnToForm(saveFormId);
                }
                handled = true;
            }
            else if (event->data.ctlSelect.controlID == b_Delete) 
            {
                // Get the database that was selected for deletion.
                int i = LstGetSelection(catNamePtr);
                Err err;

                // Confirm that the user really want's to delete a db.
                if (i != noListSelection && FrmAlert(a_AreYouSure) == 0) 
                {
                    /*
                     * Are we deleting the currently open database?
                     */
                    if ((moonPrefs.catCardNo == recordList[i]->cardNo)
                        && (catDbID == recordList[i]->dbID))
                    {
                        FrmCustomAlert(a_Warning,
                                       "Sorry. That database is in use.",
                                       "",
                                       "");
                    }
                    else 
                    {
                        err = DmDeleteDatabase(recordList[i]->cardNo,
                                               recordList[i]->dbID);
                        ErrFatalDisplayIf(err!=errNone,
                                          "Failed to delete database!");

                        /*
                         * Return to object list.
                         */
                        moonPrefs.activeFormId = saveFormId;
                        if (recordList != NULL)
                            FreeDBList(recordList,recordCount);
                        FrmReturnToForm(saveFormId);
                    }
                }
                handled = true;
            }
            else if (event->data.ctlSelect.controlID == b_Cancel)
            {
                moonPrefs.activeFormId = saveFormId;
                FrmReturnToForm(saveFormId);
                handled = true;
            }
            break;

        case frmOpenEvent:
        
            FrmDrawForm(formPtr);
            catNamePtr = (ListPtr)FrmGetObjectPtr(formPtr,
                                                  FrmGetObjectIndex(formPtr,
                                                                    ls_CatName));
            recordList = GetDBList(ASTRODBTYPE,
                                   APPID,
                                   &recordCount);
            if (recordList != NULL) 
            {
                LstSetSelection(catNamePtr,-1);
                LstSetListChoices(catNamePtr,
                                  (char**)recordList,
                                  recordCount);
                LstSetTopItem(catNamePtr,0);
                LstDrawList(catNamePtr);
            }
            
            handled = true;
            break;

        default:
            handled = false;
    }

    return handled;
}

/*
 * Events for the catalog form
 */
Boolean
CatalogFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();

    switch (event->eType) {
        case keyDownEvent:
            if (event->data.keyDown.chr == pageUpChr) 
            {
                LstScrollList(catalogPtr,winUp,8);
                handled = true;
            }
            else if (event->data.keyDown.chr == pageDownChr) 
            {
                LstScrollList(catalogPtr,winDown,8);
                handled = true;
            }
            break;
        case ctlSelectEvent:
            /*
             * Process various button clicks.
             */
            saveFormId = moonPrefs.activeFormId;
            switch (event->data.ctlSelect.controlID) {
                case s_Date5:
                    SetDate(date5Ptr);
                    break;
                case s_Time5:
                    SetTime(time5Ptr);
                    break;
                case b_Now:
                    SetNow(date5Ptr,time5Ptr);
                    break;
                case b_CatalogSelect:
                    saveFormId = moonPrefs.activeFormId;
                    FrmPopupForm(f_SelectCatalog);
                    break;
                case b_CatalogFind:
                    saveFormId = moonPrefs.activeFormId;
                    FrmPopupForm(f_FindObject);
                    break;
                case b_CatalogAgain:
                    if (catDB != NULL)
                        FindObject(catCurrRecord+1);
                    break;
            }
            handled = true;
            break;

        case lstSelectEvent:
        {
            Err err;
            
            catCurrRecord = LstGetSelection(catalogPtr);
            
            err=DmWrite(aaInfo,
                        OffsetOf(AstroAppInfoType,currRecord),
                        &catCurrRecord,
                        sizeof(UInt16));

            ErrFatalDisplayIf(err,"DMWrite Failed!");

            saveFormId = moonPrefs.activeFormId;
            FrmPopupForm(f_ObjectPosition);
            
            break;
        }
        case ctlRepeatEvent:
            if (event->data.ctlSelect.controlID == r_Dated) {
                ChangeDate(date5Ptr,-1);
            } else if (event->data.ctlSelect.controlID == r_Dateu) {
                ChangeDate(date5Ptr,1);
            }
            break;

        case frmOpenEvent:
            FrmSetMenu(formPtr,mainMenu);            
        
            FrmDrawForm(formPtr);

            date5Ptr = (ControlPtr)FrmGetObjectPtr(formPtr,
                                                   FrmGetObjectIndex(formPtr,
                                                                     s_Date5));

            time5Ptr = (ControlPtr)FrmGetObjectPtr(formPtr,
                                                   FrmGetObjectIndex(formPtr,
                                                                     s_Time5));

            CtlSetLabel(date5Ptr,
                        datestr);
            CtlSetLabel(time5Ptr,
                        timestr);

            CtlDrawControl(date5Ptr);
            CtlDrawControl(time5Ptr);

            catalogPtr = (ListPtr)FrmGetObjectPtr(formPtr,
                                                  FrmGetObjectIndex(formPtr,
                                                                    ls_Catalog));

            LoadCatalog();

            handled = true;
            break;
        
        case menuEvent:
            handled = DoMenu(event->data.menu.itemID);
            break;

        default:
            handled = false;
    }

    return handled;
}

/*
 * Release the open catalog
 */
void
FreeCat()
{
    Err err;
    
    if (catDB != NULL) 
    {
        /*
         * Close previous catalog
         */
        if (aaInfo!=NULL) 
        {
            // Unlock the app info, it doesn't
            // need to be freed.
            MemPtrUnlock(aaInfo);
            aaInfo = NULL;
        }
        
        err = DmCloseDatabase(catDB);
        ErrFatalDisplayIf(err!=errNone,
                          "Could not close catalog!");
        catDB = NULL;
    }
    
    if (catList != NULL) 
    {
        /*
         * Free previous list.
         */
        MemHandleUnlock(catList);
        MemHandleFree(catList);
        catList = NULL;
        
        MemHandleUnlock(catBuf);
        MemHandleFree(catBuf);
        catBuf = NULL;
    }    
}

/*
 * Display the currently selected object.
 */
void
LoadObject(FormPtr formPtr)
{
    double jday;
    double lstr, lsts;
    double gst, ut;
    double ra, dec;
    double alt, az;
    char   str1[FIELDLENGTH];
    int    i, h, m, s;
    
    FieldPtr objRisePtr = (FieldPtr)FrmGetObjectPtr(formPtr,
                                                    FrmGetObjectIndex(formPtr,
                                                                      f_ObjRise));
    
    FieldPtr objSetPtr = (FieldPtr)FrmGetObjectPtr(formPtr,
                                                   FrmGetObjectIndex(formPtr,
                                                                     f_ObjSet));

    MemHandle handle;
    char *ptr;
    AstroRecordType art;
    
    ErrFatalDisplayIf(catDB == NULL,"Database isn't open!");

    handle = DmQueryRecord(catDB,catCurrRecord);
    ErrFatalDisplayIf(handle == NULL,"Could not retrieve record!");

    // Unpack the record
    ptr = (char*)MemHandleLock(handle);
    StrCopy(art.description,ptr);
    ptr += StrLen(ptr); ptr++;
    memcpy((char *)&art.raH,ptr,sizeof(UInt16)*6);
    ptr += sizeof(UInt16)*6;
    StrCopy(art.magnitude,ptr);
    ptr += StrLen(ptr); ptr++;
    StrCopy(art.size,ptr);
    
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_ObjDes)),
                    art.description);
    
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_ObjMag)),
                    art.magnitude);
    
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_ObjSize)),
                    art.size);

    StrPrintF(str1,"%dh %dm %ds",
              art.raH, art.raM, art.raS);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                         FrmGetObjectIndex(formPtr,
                                                           f_ObjRA)),
                    str1);
    StrPrintF(str1,"%d\260 %d' %d\"",
              art.decD, art.decM, art.decS);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_ObjDec)),
                    str1);

    // RA is in hours, so we have to convert the hours to
    // degrees, before converting ra and dec to radians.
    ra = DegToRad(doubletime(art.raH, art.raM, art.raS)*15.0);
    dec = DegToRad(doubletime(art.decD, art.decM, art.decS));

    MemHandleUnlock(handle);
    
    jday = JDay(moonPrefs.currentDT.year,
                moonPrefs.currentDT.month,
                moonPrefs.currentDT.day,
                doubletime(moonPrefs.currentDT.hour,
                           moonPrefs.currentDT.minute,0),
                moonPrefs.tz);

    gst = JDayToGST(jday);
    lstr = GSTToLST(gst,moonPrefs.lng);

    RaDecToAltAz(ra, dec,
                 lstr, moonPrefs.lat,
                 &alt, &az);
    
    RadToDMS(alt,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                                f_ObjAlt)),
                    str1);
    
    RadToDMS(az,
             &h,&m,&s);
    StrPrintF(str1,"%d\260 %d' %d\"",
              h,m,s);
    UpdateTextField((FieldPtr)FrmGetObjectPtr(formPtr,
                                              FrmGetObjectIndex(formPtr,
                                                           f_ObjAz)),
                    str1);

    i = RiseSet(ra,dec,
                moonPrefs.lat,
                &lstr, &lsts);
    if (i) 
    {
        // Object never rises or sets. It's either always
        // above or below the horizon.
        UpdateTextField(objRisePtr,"--h --m --s");
        UpdateTextField(objSetPtr,"--h --m --s");
    }
    else 
    {
        gst = LSTToGST(lstr,moonPrefs.lng);
        ut = GSTToUT(gst,
                     moonPrefs.currentDT.year,
                     moonPrefs.currentDT.month,
                     moonPrefs.currentDT.day);
        UpdateTimeField(objRisePtr,UTToLT(ut,moonPrefs.tz),0);

        gst = LSTToGST(lsts,moonPrefs.lng);
        ut = GSTToUT(gst,
                     moonPrefs.currentDT.year,
                     moonPrefs.currentDT.month,
                     moonPrefs.currentDT.day);
        UpdateTimeField(objSetPtr,UTToLT(ut,moonPrefs.tz),0);
    }
}

/*
 * Display the current catalog of objects.
 */
void
LoadCatalog() 
{
    LocalID local;
    MemHandle handle;
    UInt16 i, j;
    UInt16 offset;
    Err err;
    AstroRecordType* art;
    char *catPtr;

    /*
     * Erase the current list
     */
    LstSetListChoices(catalogPtr,
                      NULL,NULL);

    FreeCat();
    
    if (catDbID == 0) 
    {
        // See if we can find the last used catalog.
        catDbID = DmFindDatabase(moonPrefs.catCardNo,
                                 moonPrefs.catDbName);
        if (catDbID == 0)
            return;
    }

    catDB = DmOpenDatabase(moonPrefs.catCardNo,
                           catDbID,
                           dmModeReadWrite);
    ErrFatalDisplayIf(catDB==0,
                      "Could not open the database!");

    /*
     * Get the application info.
     */
    err = DmDatabaseInfo(moonPrefs.catCardNo,
                         catDbID,
                         NULL, NULL, NULL, NULL,
                         NULL, NULL, NULL,
                         &local,
                         NULL, NULL, NULL);
    ErrFatalDisplayIf(err!=errNone,
                      "Could not retrieve catalog info.");

    aaInfo = MemLocalIDToLockedPtr(local,
                                   moonPrefs.catCardNo);
    /*
     * Get the cat buffer.
     */
    catBuf = MemHandleNew(sizeof(art->description)*
                          aaInfo->numRecords);
    ErrFatalDisplayIf(catBuf==NULL,
                      "Could not allocate enough memory.");

    catPtr = MemHandleLock(catBuf);

    /*
     * Get the cat strings
     */

    j = 0; offset = 0;
    for(i=0;i<aaInfo->numRecords;i++) 
    {
        handle = DmQueryRecord(catDB,i);
        art = (AstroRecordType*)MemHandleLock(handle);
        StrCopy(catPtr+offset,art->description);
        offset += StrLen(art->description)+1;
        MemHandleUnlock(handle);
    }

    MemHandleUnlock(catBuf);
    MemHandleResize(catBuf,offset);
    MemHandleLock(catBuf);
    
    catList = SysFormPointerArrayToStrings(catPtr,
                                           aaInfo->numRecords);
    ErrFatalDisplayIf(catList == NULL,
                      "Failed to build catalog list.");

    LstSetListChoices(catalogPtr,
                      MemHandleLock(catList),
                      aaInfo->numRecords);
    LstSetTopItem(catalogPtr,aaInfo->currRecord);
    FrmUpdateForm(f_Catalog,frmRedrawUpdateCode);
}

/*
 * Display the current catalog of objects.
 */
void
FindObject(UInt16 start) 
{
    Err err;
    MemHandle handle;
    UInt16 i;
    AstroRecordType* art;
    char description[sizeof(art->description)];
    
    i = start;
    while (i < aaInfo->numRecords)
    {
        handle = DmQueryRecord(catDB,i);
        ErrFatalDisplayIf(handle==NULL,"DmQueryRecord failed!");
        art = (AstroRecordType*)MemHandleLock(handle);
        StrToLower(description,art->description);
        if (StrStr(description,findString) != NULL)
        {
            MemHandleUnlock(handle);
            LstSetTopItem(catalogPtr,i);
            LstSetSelection(catalogPtr,i);
            catCurrRecord = i;
            
            err=DmWrite(aaInfo,
                        OffsetOf(AstroAppInfoType,currRecord),
                        &catCurrRecord,
                        sizeof(UInt16));

            ErrFatalDisplayIf(err,"DMWrite Failed!");
            break;
        }
        MemHandleUnlock(handle);
        i++;
    } 

    SndPlaySystemSound(sndInfo);
    
    FrmUpdateForm(f_Catalog,frmRedrawUpdateCode);
}
