/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: AstroLib.c,v 1.11 2000/09/28 03:04:29 mheinz Exp $
 * Copyright (C) 2000, Michael Heinz
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include "MathLib.h"
#include "PalmUtil.h"
#include "constants.h"
#include "AstroLib.h"
#include "resource.h"

#ifdef QUADATAN
/*
 * ATAN(x/y) that has been corrected for 
 * quadrant ambiguities caused by the signs
 * of X and Y.
 *
 * MHEINZ: Replaced this function with MathLib's atan2() function.
 */
double
QuadATan(double y,
         double x)
{

    if (x > 0) 
        return atan(y/x);
    else if (x < 0)
        return atan(y/x) + PI;
    else if (y > 0)
        return PIh;
    else if (y < 0)
        return -PIh;
    else
        return 0.0;
}
#endif

/*
 * Normalizes a radian angle to the range 0 - 2PI
 */
inline double
RadScale(double d)
{
    d = fmod(d,PI_2);
    if (d < 0.0)
        d += PI_2;
    
    return d;
}

/*
 * Normalizes a degree angle to the range 0-360.
 */
double
DegScale(double d)
{
    d = fmod(d,360.0);
    if (d < 0.0)
        d += 360.0;
    
    return d;
}

/*
 * Scale a value from 0 - 1
 */
double
Normalize(double d)
{
    d = d - trunc(d);

    if (d < 0)
        d += 1.0;

    return d;
}

/*
 * Returns the Julian Day.
 *
 * This formula is valid for dates from 1901 to 2099
 */
double
JDay(int year, int month, int day,
     double local_time,
     double tz)
{
    double A, B, C, D, F;
    
    if (month<3) {
        month += 12;
        year -= 1;
    }

    if (year>=1582) {
        A = trunc(year / 100);
        B = 2 - A + trunc(A/4);
    } else {
        B = 0.0;
    }

    if (year < 0)
        C = trunc((365.25 * year) - 0.75);
    else
        C = trunc(365.25 * year);

    D = trunc(30.6001 * (month + 1));

    F = B + C + D + day + 1720994.5 + (local_time - tz)/24.0;

    return F;
}

/*
 * Calculate the phase of the moon and % full for the current julian
 * day.
 */
void
LunarPhase(double *age,
           double *full,
           double jday)
{
    *age = (jday - 2451550.1) / 29.530588853;
    *age = Normalize(*age);

    if (*age < 0.5)
        *full = *age / 0.5;
    else
        *full = (1 - *age) / 0.5;

    return;
}

/*
 * Convert a Julian Day to GST
 */
double
JDayToGST(double jday)
{
    double T, T0, G, UT;
    double jd, jf;

    /*
     * We need the julian day for midnight UT,
     * and the UT since midnight.
     */
    jd = trunc(jday);
    jf = jday - jd;
    if (jf < 0.5) {
        jd -= 0.5;
        jf += 0.5;
    } else {
        jd += 0.5;
        jf -= 0.5;
    }

    UT = jf * 24.0;
    
    T = (jd - 2451545.0) / 36525.0;

    T0 = 6.697374558 + (2400.051336 * T) + (0.000025862 * T * T);

    T0 = fmod(T0,24.0);
    if (T0 < 0.0)
        T0 += 24.0;
    
    G = UT * 1.002737909 + T0;
    G = fmod(G,24.0);
    if (G < 0.0)
        G += 24.0;
    
    return G;
}

/*
 * Convert a gst and longitude to LST.
 * longitude is in radians.
 */
double
GSTToLST(double g,
         double lng)
{
    double l;
        
    l = g - (lng / DR / 15.0);
    return l;
}

/*
 * Convert a LST and longitude to GST
 * longitude is in radians.
 */
double
LSTToGST(double l,
         double lng)
{
    double g;
    
    g = l + (lng / DR / 15.0);
    return g;
}

/*
 * Convert degrees to radians
 */
double
DegToRad(double deg)
{
    return deg * DR;
}

/*
 * Convert radians to degrees
 */
double 
RadToDeg(double rad)
{
    return RadScale(rad) / DR;
}

/*
 * Convert radians to degrees, minutes, seconds
 *
 * Note that for negative values of rad, only h will be negative in
 * the result. (i.e., -0.79419... radians = -45, 30, 15)
 */
void
RadToDMS(double rad,
         int *h,
         int *m,
         int *s)
{
    int sign = (rad < 0);
    double deg = fabs(rad / DR);
    
    *h = trunc(deg);
    deg = (deg - *h) * 60.0;
    
    *m = trunc(deg);
    deg = (deg - *m) * 60.0;
    
    *s = trunc(deg);

    if (sign)
        *h = -(*h);
}

/*
 * Convert DMS to decimal degrees
 *
 * Note that the assumption is that for negative values of DMS, only d
 * is actually negative.
 */
double
DMSToDeg(int d,
         int m,
         int s)
{
    double t;
    if (d < 0) {
        /* Make sure all signs match */
        m = -m;
        s = -s;
    }
    t = d + (m / 60.0) + (s / 3600.0);

    return t;
}

/*
 * Convert degrees to degrees, minutes, seconds
 *
 * Note that for negative values of deg, only h will be negative in
 * the result.
 */
void
DegToDMS(double deg,
         int *h,
         int *m,
         int *s)
{
    int sign = (deg < 0);

    deg = fabs(deg);
    *h = trunc(deg);
    deg = (deg - *h) * 60.0;
    
    *m = trunc(deg);
    deg = (deg - *m) * 60.0;
    
    *s = trunc(deg);
    if (sign)
	*h = -(*h);
}

/*
 * Convert radians to hours, minutes, seconds
 *
 * Note that for negative values of rad, only h will be negative in
 * the result.
 */
void
RadToHMS(double rad,
         int *h,
         int *m,
         int *s)
{
    int sign = (rad < 0);
    double deg = fabs(rad / DR / 15.0);
    
    *h = trunc(deg);
    deg = (deg - *h) * 60.0;
    
    *m = trunc(deg);
    deg = (deg - *m) * 60.0;
    
    *s = trunc(deg);
    if (sign)
        *h = -(*h);
}

/*
 * Convert degrees to hours, minutes, seconds
 *
 * Note that for negative values of deg, only h will be negative in
 * the result. (i.e., -45.5041666... degrees = -3, 6, 0)
 */
void
DegToHMS(double deg,
         int *h,
         int *m,
         int *s)
{
    int sign = (deg < 0);
    deg = fabs(deg / 15.0);
    
    *h = trunc(deg);
    deg = (deg - *h) * 60.0;
    
    *m = trunc(deg);
    deg = (deg - *m) * 60.0;
    
    *s = trunc(deg);
    if (sign)
        *h = -(*h);
}

/*
 * Convert R.A./Dec to Alt/Az
 *
 * All angles in RADIANS.
 */
void
RaDecToAltAz(double ra,
             double dec,
             double lst,
             double lat,
             double *alt,
             double *az)
{
    double      HourAngle = (lst * DR * 15.0) - ra;

    double      sina, cosA;
     
    sina = sin(dec) * sin(lat) + cos(dec)*cos(lat)*cos(HourAngle);
    *alt = asin(sina);
    
    cosA = (sin(dec) - sin(lat)*sina) / (cos(lat) * cos(*alt));

    *az = acos(cosA);

    if (sin(HourAngle) > 0 )
        *az = PI_2 - *az;
}

/*
 * Convert celestial co-ordinates (lat,long)
 * to right ascension and declination.
 *
 * All angles in radians.
 */
void
CelestialToRaDec(double lat,
                 double lng,
                 double jday,
                 double *ra,
                 double *dec)
{
    double      x, y;
    double      sind;
    double      eo, t;

    /*
     * Calculate ecliptic obliquity
     */
    t = (jday - 2451545.0)/36525.0; /* EPOCH = January 2000 1.5 */
    eo = .40909281392 +
         ((-46.815 * t - 0.0006 * t * t + 0.00181 * t * t * t)/3600.0 * DR);
    
    sind = sin(lat)*cos(eo)+cos(lat)*sin(eo)*sin(lng);
    *dec = asin(sind);

    y = sin(lng)*cos(eo) - tan(lat)*sin(eo);
    x = cos(lng);

#ifdef QUADATAN    
    *ra = QuadATan(y,x);
#else
    *ra = RadScale(atan2(y,x));
    
#endif
}

/*
 * Solve Kepler's Equation.  All units in Radians.
 */
double
Kepler(double MeanAnomoly,
       double Eccentricity)
{
    double      E, dE, d;

    E = MeanAnomoly;
    d = E - Eccentricity*sin(E) - MeanAnomoly;

    while (fabs(d) > 0.000001) {
        dE = d / (1 - Eccentricity*cos(E));
        E = E - dE;
        d = E - Eccentricity*sin(E) - MeanAnomoly;
    }
    return E;
}

/*
 * Return the longitude of the Sun for a given julian day
 * If Ms is not null, return the mean anomoly in Ms.
 */
double
SolarLong(double jday,
          double *Ms)
{
    double N;   /* # of day/degrees since the epoch */
    double M;   /* mean anomoly */
    double E,v,tv2;  /* Error Correction */
    double l;   /* longitude of the sun */

    /* Angular distance since the Epoch began. */
    N = SOLAR_RpD * (jday - EPOCH);
    
    /*
     * The 'mean anomoly' - an error correction to compensate for the
     * irregularity of the Earth's orbit around the sun.
     */
    M = N + SOLAR_EG - SOLAR_MG;

    if (Ms) {
        /* Return the mean anomoly */
        *Ms = fmod(M,PI_2);
    }

    /*
     * Calculate the true anomoly
     */
    E = Kepler(M,SOLAR_E);

    tv2 = sqrt((1+SOLAR_E)/(1-SOLAR_E))*tan(E/2.0);

    v = atan(tv2) * 2;

    /*
     * Calculate the longitude.
     */
    l = RadScale(v + SOLAR_MG);
    
    return l;
}

/*
 * Calculate the rising and setting times for an object.
 *
 * Times in LST.  If the object does not rise/set at this
 * latitude, returns -1, and LSTr and LSTs are not set.
 *
 * ra and dec are in radians.
 */
int
RiseSet(double ra,
        double dec,
        double lat,
        double *LSTr,
        double *LSTs)
{
    double      tt, H, rah;
    
    tt = -tan(lat) * tan(dec);
    
    if (tt > 1.0)
        return -1;
    else if (tt < -1.0)
        return 1;

    /*
     * Hour angle
     */
    H = acos(tt)/15.0/DR;

    rah = ra/15.0/DR;
    
    *LSTr = 24 - H + rah;
    *LSTs = rah + H;

    while (*LSTr > 24.0)
        *LSTr -= 24.0;
    
    while (*LSTs > 24.0)
        *LSTs -= 24.0;

#ifdef RISESET_CHECK
    {
        char s2[32], s3[32];
        int h,m,s;
        
        DegToDMS(*LSTr,
                 &h,&m,&s);
        StrPrintF(s2,"%dh %dm %ds",
                  h,m,s);

        DegToDMS(*LSTs,
                 &h,&m,&s);
        StrPrintF(s3,"%dh %dm %ds",
                  h,m,s);
        FrmCustomAlert(a_Debug,"Rise/Set",s2,s3);        
    }
#endif

    return 0;
}

/*
 * Convert GST to UT.
 */
double
GSTToUT(double gst,
        int year,
        int month,
        int day)
{
    double jday;
    double S, T, T0, UT;
    int df;
    
    jday = JDay(year,month,day,
                0.0,0.0);

    S = jday - 2451545.0;
    
    T = S / 36525.0;

    T0 = 6.697374558 + (2400.051336 * T) + (0.000025862 * T * T);

    T0 = fmod(T0,24.0);
    if (T0 < 0.0)
        T0 += 24.0;
    
    gst -= T0;

    /*
     * Remeber if we crossed the midnight line.
     */
    df = (gst > 24.0);
    
    gst = fmod(gst,24.0);
    if (gst < 0.0)
        gst += 24.0;

    /*
     * The local day is 4 minutes shorter than the sidereal day, so we
     * multiply by that ratio.  If the gst was greater than 24 before
     * we modulo'ed it, add 24 hours.
     */
    UT = gst * 0.9972695663;
    if (df)
        UT +=24.0;
    
    return UT;
}

/*
 * Returns the times of sunrise and sunset in UT.
 */
int
SunRiseSet(int year,
           int month,
           int day,
           double tz,
           double lat,
           double lng,
           double *UTr,
           double *UTs)
{
    double gstr, gsts;
    double jday;
    double lstr, lsts;
    double ra, dec, l;
    double y;
    double w, dt;
    int i;

    /*
     * Local Noon.
     */
    jday = JDay(year,month,day,12.0,tz);
    
    /*
     * Longitude of the sun
     */
    l = SolarLong(jday,
                  NULL);

    /*
     * ra and dec for noon.
     */
    CelestialToRaDec(0.0,l,
                     jday,
                     &ra, &dec);

    i = RiseSet(ra,
                dec,
                lat,
                &lstr,
                &lsts);

    if (i)
        return i;

    gstr = LSTToGST(lstr,lng);
    gsts = LSTToGST(lsts,lng);

    w = acos(sin(lat)/cos(dec));
    
    y = asin(sin(SOLAR_A / 2.0 - SOLAR_P + IOR_EST)/sin(w));
    dt = 0.066666666667 * y * cos(dec) / DR;
    
    /*
     * Sunrise
     */
    *UTr = GSTToUT(gstr - dt,year,month,day);
    
    /*
     * Sunset
     */
    *UTs = GSTToUT(gsts + dt,year,month,day);

    return 0;
}

/*
 * Lunar Position for the current date/time angles in radians.
 *
 * ra2, dec2 and EC are optional returns.  EC is the (optional) error
 * correcting term for parallax equations. If EC = NULL then it is not
 * populated.
 */
void
LunarPos(double *ra,
         double *dec,
         double jday,
         double *EC)
{
    double ls, ms;                      /* solar info */
    double lm, mm, nm;                  /* lunar info */
    double ev, ae, a3, a4, ec, v;       /* various correcting factors */
    double mlat, mlng;                  /* moon's pos */
    double x, y;
    
    /*
     * Longitude and anomoly of the sun
     */
    ls = SolarLong(jday,
                   &ms);

    /*
     * Mean Longitude and anomoly of the moon
     */
    lm = 13.176396*(jday-EPOCH)*DR + MOON_L0;
    mm = lm - 0.1114041*(jday-EPOCH)*DR - MOON_P0;

    /*
     * Node longitude
     */
    nm = MOON_N0 - 0.0529539*(jday - EPOCH)*DR - 0.16*sin(ms)*DR;
    
    /*
     * Corrections
     */
    ev = 1.2739 * sin(2 *(lm - ls) - mm) * DR;
    ae = 0.1858 * sin(ms) * DR;
    a3 = 0.37 * sin(ms) * DR;
    
    /*
     * Corrected anomoly
     */
    mm = mm + ev - ae - a3;

    /*
     * Equation of the center
     */
    ec = 6.2886 * sin(mm) * DR;

    if (EC)
        *EC = mm + ec;
    
    /*
     * Y.A.C.
     */
    a4 = 0.214 * sin(2 * mm) * DR;

    lm = lm + ev + ec - ae + a4;

    v = 0.6583 * sin(2 * (lm - ls)) * DR;

    lm = lm + v;

    /*
     * Moon's position
     */
    mlat = asin(sin(lm - nm)*sin(MOON_I));

    x = cos(lm - nm);
    y = sin(lm - nm) * cos(MOON_I);

#if QUADATAN
    mlng = QuadATan(y,x);
#else
    mlng = RadScale(atan2(y,x));
#endif
    
    mlng += nm;
    
    /*
     * ra and dec for midnight
     */
    CelestialToRaDec(mlat, mlng,
                     jday,
                     ra, dec);
}

/*
 * Returns the times of moon rise and moon set in UT.
 */
int
MoonRiseSet(int year,
            int month,
            int day,
            double tz,
            double lat,
            double lng,
            double *UTr,
            double *UTs)
{
    double gstr, gsts;
    double jday, a;
    double lstr, lsts;
    double ra, dec, EC;
    double y;
    double px, rho, w, dt;
    double UTtemp;
    int i;

    /*
     * Local noon.
     */
    jday = JDay(year,month,day,
                12.0,tz);

    /*
     * First Approximation of position
     */
    LunarPos(&ra, &dec,
             jday,
             &EC);
    
    /*
     * Initial guess at rise and set times.
     */
    i = RiseSet(ra,
                dec,
                lat,
                &lstr,
                &lsts);
 
    if (i)
        return i;

    gstr = LSTToGST(lstr,lng);    
    gsts = LSTToGST(lsts,lng);

    rho = (1 - MOON_E*MOON_E) / ( 1 + MOON_E * cos(EC));

    px = MOON_E / rho;

    w = acos(sin(lat)/cos(dec));
    y = asin(sin(MOON_A / 2.0 - px + IOR_EST)/sin(w));
    dt = 0.066666666667 * y * cos(dec) / DR;

    /*
     * First approximation of rise and set.
     */
    *UTr = GSTToUT(gstr - dt,year,month,day);
    *UTs = GSTToUT(gsts + dt,year,month,day);

    UTtemp = *UTr;

    do
    {
#ifdef MOONRISE_CHECK
        {
            char s1[32], s2[32], s3[32];
            int h,m,s;
            
            DegToDMS(dt,
                     &h,&m,&s);
            StrPrintF(s1,"dt: %dh %dm %ds",
                      h,m,s);
            
            DegToDMS(*UTr+tz,
                     &h,&m,&s);
            StrPrintF(s2,"R1: %dh %dm %ds",
                      h,m,s);
            
            DegToDMS(UTtemp+tz,
                     &h,&m,&s);
            StrPrintF(s3,"R2: %dh %dm %ds",
                      h,m,s);
            FrmCustomAlert(a_Debug,s1,s2,s3);        
        }
#endif
        
        /*
         * Convert approximated rise time to JDAY.
         * Adjust julian day if local date isn't the same
         * as Greenwich date.
         */
        jday = JDay(year,month,day,*UTr,0.0);
        a = *UTr + tz;
        if (a < 0)
            jday += 1.0;
        else if (a > 24.0)
            jday -= 1.0;
        
        /*
         * Improved position at rise time.
         */
        LunarPos(&ra, &dec,
                 jday,
                 &EC);
    
        /*
         * Improved guess at rise time.
         * We ignore the set time.
         */
        i = RiseSet(ra,
                    dec,
                    lat,
                    &lstr,
                    &lsts);
 
        if (i)
            return i;

        gstr = LSTToGST(lstr,lng);
    
        rho = (1 - MOON_E*MOON_E) / ( 1 + MOON_E * cos(EC));

        px = MOON_E / rho;

        w = acos(sin(lat)/cos(dec));
        y = asin(sin(MOON_A / 2.0 - px + IOR_EST)/sin(w));
        dt = 0.066666666667 * y * cos(dec) / DR;

        if (UTtemp != -100.0)
            *UTr = UTtemp;
        UTtemp = GSTToUT(gstr - dt,year,month,day);
    }
    while (fabs(*UTr - UTtemp) > .05);

#ifdef MOONRISE_CHECK
    {
        char s2[32], s3[32];
        int h,m,s;
        
        DegToDMS(*UTr+tz,
                 &h,&m,&s);
        StrPrintF(s2,"R1: %dh %dm %ds",
                  h,m,s);

        DegToDMS(UTtemp+tz,
                 &h,&m,&s);
        StrPrintF(s3,"R2: %dh %dm %ds",
                  h,m,s);
        FrmCustomAlert(a_Debug,"Final Rise",s2,s3);        
    }
#endif
    
    *UTr = UTtemp;
    UTtemp = *UTs;

    do
    {
#ifdef MOONRISE_CHECK
        {
            char s2[32], s3[32];
            int h,m,s;
            
            DegToDMS(*UTs+tz,
                     &h,&m,&s);
            StrPrintF(s2,"S1: %dh %dm %ds",
                      h,m,s);
            
            DegToDMS(UTtemp+tz,
                     &h,&m,&s);
            StrPrintF(s3,"S2: %dh %dm %ds",
                      h,m,s);
            FrmCustomAlert(a_Debug,"Set Test",s2,s3);        
        }
#endif
        /*
         * Convert approximated set time to JDAY.
         * Adjust julian day if local date isn't the same
         * as Greenwich date.
         */
        jday = JDay(year,month,day,*UTs,0.0);
        a = *UTs + tz;
        if (a < 0)
            jday += 1.0;
        else if (a > 24.0)
            jday -= 1.0;
        
        /*
         * Improved position at set time.
         */
        LunarPos(&ra, &dec,
                 jday,
                 &EC);
    
        /*
         * Improved guess at set time
         * We ignore the rise time.
         */
        i = RiseSet(ra,
                    dec,
                    lat,
                    &lstr,
                    &lsts);
        
        if (i)
            return i;

        gsts = LSTToGST(lsts,lng);
        
        rho = (1 - MOON_E*MOON_E) / ( 1 + MOON_E * cos(EC));

        px = MOON_E / rho;

        w = acos(sin(lat)/cos(dec));
        y = asin(sin(MOON_A / 2.0 - px + IOR_EST)/sin(w));
        dt = 0.066666666667 * y * cos(dec) / DR;

        *UTs = UTtemp;
        UTtemp = GSTToUT(gsts + dt,year,month,day);
    }
    while (fabs(*UTs - UTtemp) > .05);
    
#ifdef MOONRISE_CHECK
    {
        char s2[32], s3[32];
        int h,m,s;
        
        DegToDMS(*UTs+tz,
                 &h,&m,&s);
        StrPrintF(s2,"S1: %dh %dm %ds",
                  h,m,s);

        DegToDMS(UTtemp+tz,
                 &h,&m,&s);
        StrPrintF(s3,"S2: %dh %dm %ds",
                  h,m,s);
        FrmCustomAlert(a_Debug,"Final Set",s2,s3);        
    }
#endif
    *UTs = UTtemp;

    return 0;
}

/*
 * UT to local time (decimal time)
 *
 * This routine ensures that the return value is > 0.0 (i.e., if (ut -
 * tz) < 0 then add 24.0.)
 *
 * tz = time zone, in hours east of UT (i.e., EST = -5.0)
 */
double UTToLT(double ut,
              double tz)
{
    ut += tz;
    if (ut < 0.0)
        ut += 24.0;

    return ut;
}

/*
 * Data for planetary equations.
 * see astrolib.h for definitions of MERCURY, MARS, etc..
 */

/* Period of orbit (years) */
static double period[] = { 0.240852,
                           0.615211,
                           1.00004,
                           1.880932,
                           11.863075,
                           29.471362,
                           84.039492,
                           164.79246,
                           246.77027};

/* Longitude at epoch (Degrees)*/
static double elong[] = { 60.750646,
                          88.455855,
                          99.403308,
                          240.739474,
                          90.638185,
                          287.690033,
                          271.063148,
                          282.349556,
                          221.4127};

/* longitude at perihelion (degrees) */
static double plong[] = { 77.299833,
                          131.430236,
                          102.468413,
                          335.874939,
                          14.170747,
                          92.861407,
                          172.884833,
                          48.009758,
                          224.133};

/* Eccentricity of the orbit */
static double ecc[] = { 0.205633,
                        0.006778,
                        0.016713,
                        0.093396,
                        0.048482,
                        0.055581,
                        0.046321,
                        0.009003,
                        0.24624};

/* Semi-major axis of the orbit (AU's) */
static double sma[] = { 0.387099,
                        0.723332,
                        1.0, 
                        1.523688,
                        5.202561,
                        9.554747,
                        19.21814,
                        30.109570,
                        39.3414};

/* inclination of the orbit (degrees) */
static double inc[] = { 7.004540,
                        3.394535,
                        0.0,
                        1.849736,
                        1.303613,
                        2.488980,
                        0.773069,
                        1.770646,
                        17.1420};

/* longitude of the ascending node (Degrees) */
static double asc[] = { 48.212740,
                        76.589820,
                        0.0,
                        49.480308,
                        100.353142,
                        113.576139,
                        73.926961,
                        131.670599,
                        110.144};

#if 0
/* angular diameter @ 1 AU (arc seconds)*/
static double adau[] = { 6.74,
                         16.92,
                         0.0,  
                         9.36,
                         196.74,
                         165.60,
                         65.80,
                         62.20,
                         8.20};

/* Visual magnitude */
static double vmau[] = { -0.42,
                         -4.40,
                         0.0,  
                         -1.52,
                         -9.40,
                         -8.88,
                         -7.19,
                         -6.87,
                         -1.0};
#endif

#define PLUTOEPOCH      2447161.5

/*
 * Calculate the position of a planet for a given
 * julian date.
 */
void
PlanetPos(double *ra,
          double *dec,
          int planet,
          double jday)
{
    double ma, v, d, l, r, hl, e, tv2;
    double Ema, Ev, El, Er;
    double x, y;
    double lat, lng;
    
    if (planet != PLUTO)
        /*
         * days since 1990
         */
        d = jday - EPOCH;
    else
        /*
         * days since 1988
         */
        d = jday - PLUTOEPOCH;

    /*
     * Longitude for target planet (in degrees)
     */
    l = 360.0/365.242191 * d / period[planet];
    
    /*
     * Mean Anomoly
     */
    ma = DegScale(l + elong[planet] - plong[planet]);
    
    e = Kepler(ma * DR,ecc[planet]);
    tv2 = sqrt((1+ecc[planet])/(1-ecc[planet]))*tan(e/2);
    
    /*
     * True Anomoly
     */
    v = RadScale(atan(tv2) * 2);
    
    l = v + (plong[planet] * DR);

    r = sma[planet]*(1 - (ecc[planet] * ecc[planet]))/
        (1 + ecc[planet]*cos(v));

    /*
     * Longitude for Earth
     */
    El = 360/365.242191 * d / period[EARTH];

    /*
     * Mean Anomoly
     */
    Ema = DegScale(El + elong[EARTH] - plong[EARTH]);

    e = Kepler(Ema * DR,ecc[EARTH]);
    tv2 = sqrt((1+ecc[EARTH])/(1-ecc[EARTH]))*tan(e/2);

    /*
     * True Anomoly
     */
    Ev = RadScale(atan(tv2) * 2);

    El = Ev + (plong[EARTH] * DR);

    Er = sma[EARTH]*(1 - (ecc[EARTH] * ecc[EARTH]))/
        (1 + ecc[EARTH]*cos(Ev));
    
    /*
     * Heliocentric latitude of planet
     */
    hl = asin(sin(l - (asc[planet]*DR))*sin(inc[planet]*DR));

    y = sin(l - (asc[planet]*DR))*cos(inc[planet]*DR);
    x = cos(l - (asc[planet]*DR));

#ifdef QUADATAN    
    l = QuadATan(y,x);
#else
    l = RadScale(atan2(y,x));
#endif
    
    l += asc[planet]*DR;
    
    r = r * cos(hl);

    if (planet < EARTH) {
        lng = PI + El + atan((r*sin(El - l))/(Er - r*cos(El - l)));
    } else {
        lng = atan((Er*sin(l - El))/(r - Er*cos(l - El))) + l;
    }
    
    lat = atan((r*tan(hl)*sin(lng-l))/
               (Er*sin(l-El)));
    
    CelestialToRaDec(lat,
                     lng,
                     jday,
                     ra,
                     dec);
    return;
}

/*
 * Returns the times of planet rise and planet set in UT.
 */
int
PlanetRiseSet(int planet,
              int year,
              int month,
              int day,
              double lat,
              double lng,
              double *UTr,
              double *UTs)
{
    double      ra, dec;
    double      jday;
    int         i;
    double      lstr, lsts;
    double      gstr, gsts;
    double      y, w, dt, px;
    
    jday = JDay(year,month,day,0.0,0.0);

    PlanetPos(&ra,&dec,planet,jday);

    i = RiseSet(ra,dec,lat,
                &lstr, &lsts);
    if (i)
        return i;

    gstr = LSTToGST(lstr,lng);
    gsts = LSTToGST(lsts,lng);
    
    /* Time shift for parallax and refraction */
    px = 0.0004263452 / sma[planet]; /* parallax */
    w = acos(sin(lat)/cos(dec));
    y = asin(sin(IOR_EST - px)/sin(w)); 
    dt = 0.066666666667 * y * cos(dec) / DR;

    *UTr = GSTToUT(gstr - dt,year,month,day);
    *UTs = GSTToUT(gsts + dt,year,month,day);

    return 0;
}
