/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: AstroLib.h,v 1.8 2000/09/27 01:43:17 mheinz Exp $
 * Copyright (C) 2000, Michael Heinz
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef AstroLib_h
#define AstroLib_h

/*
 * Scale a value from 0 - 1
 */
double Normalize(double d);

/*
 * Returns the Julian Day.
 *
 * This formula is valid for dates from 1901 to 2099
 */
double JDay(int year, int month, int day,
            double local_time, double tz);

/*
 * Convert a Julian day to GST
 */
double JDayToGST(double jday);

/*
 * Convert a gst and longitude to LST
 * longitude is in radians.
 */
double GSTToLST(double gst, double lng);

/*
 * Convert a LST and longitude to GST.
 * longitude is in radians.
 */
double LSTToGST(double lst, double lng);

/*
 * Normalizes a radian angle to the range 0 - 2PI
 */
double RadScale(double d);

/*
 * Normalizes a degree angle to the range 0-360.
 */
double DegScale(double d);

/*
 * Convert degrees to radians
 */
double DegToRad(double deg);

/*
 * Convert radians to degrees
 */
double RadToDeg(double rad);

/*
 * Convert radians to degrees, minutes, seconds
 *
 * Note that for negative values of rad, only h will be negative in
 * the result. (i.e., -0.79419... radians = -45, 30, 15)
 */
void RadToDMS(double rad,
              int *h,
              int *m,
              int *s);

/*
 * Convert radians to hours, minutes, seconds
 */
void RadToHMS(double rad,
              int *h,
              int *m,
              int *s);
/*
 * Convert decimal degrees to hours, minutes, seconds
 *
 * Note that for negative values of deg, only h will be negative in
 * the result. (i.e., -45.5041666... degrees = -45, 30, 15)
 */
void DegToDMS(double deg,
              int *h,
              int *m,
              int *s);

/*
 * Convert decimal degrees to hours, minutes, seconds
 *
 * Note that for negative values of deg, only h will be negative in
 * the result. (i.e., -45.5041666... degrees = -3, 6, 0)
 */
void DegToHMS(double deg,
              int *h,
              int *m,
              int *s);

/*
 * UT to local time (decimal time)
 *
 * This routine ensures that the return value is > 0.0 (i.e., if (ut -
 * tz) < 0 then add 24.0.)
 *
 * tz = time zone, in hours east of UT (i.e., EST = -5.0)
 */
double UTToLT(double ut,
              double tz);

/*
 * Convert GST to UT.
 */
double GSTToUT(double gst,
               int year,
               int month,
               int day);

/*
 * Convert DMS to decimal degrees. Note that for negative values, the
 * assumption is that d is negative, but m and s are always positive.
 */
double DMSToDeg(int d,
                int m,
                int s);

/*
 * Convert R.A./Dec to Alt/Az
 *
 * ra = Right Ascension, dec = Declination
 * lst = observer's Local Sidereal Time, lat = observer's Latitude
 * alt = the final altitude, az = the final azimuth.
 *
 * All angles in RADIANS.
 */
void RaDecToAltAz(double ra, double dec,
                  double lst, double lat,
                  double *alt, double *az);

/*
 * Convert celestial co-ordinates (lat,long)
 * to right ascension and declination.
 *
 * All angles in radians.
 */
void CelestialToRaDec(double lat,
                      double lng,
                      double jday,
                      double *ra,
                      double *dec);
/*
 * Return the longitude of the Sun for a given julian day.
 * If ms is !NULL then returns the mean anomoly, too.
 */
double SolarLong(double jday, double *ms);

/*
 * Calculate the rising and setting times for an object.
 *
 * Times in LST.  If the object does not rise/set at this
 * latitude, returns -1, and LSTr and LSTs are not set.
 *
 * ra and dec are in radians.
 */
int RiseSet(double ra,
            double dec,
            double lat,
            double *LSTr,
            double *LSTs);

/*
 * Calculate the rising and setting times for the sun.
 *
 * Times in UT.  If the object does not rise/set at this
 * latitude, returns -1, and UTr and UTs are not set.
 *
 * lat and lng are the observer's position in radians.
 */
int SunRiseSet(int year,
               int month,
               int day,
               double tz,
               double lat,
               double lng,
               double *UTr,
               double *UTs);

/*
 * Calculate the rising and setting times for the moon.
 *
 * Times in UT.  If the object does not rise/set at this
 * latitude, returns -1, and UTr and UTs are not set.
 *
 * lat and lng are the observer's position in radians.
 */
int MoonRiseSet(int year,
                int month,
                int day,
                double tz,
                double lat,
                double lng,
                double *UTr,
                double *UTs);

/*
 * Calculate the phase of the moon and % full for the current julian
 * day.
 */
void LunarPhase(double *age,
                double *full,
                double jday);

/*
 * Lunar Position for the current date/time
 * angles in radians.  If EC is non-null
 * it will contain an estimated correction
 * for parallax.
 */
void LunarPos(double *ra,
              double *dec,
              double jday,
              double *EC);

/*
 * Planet numbers
 */
#define MERCURY         0
#define VENUS           1
#define EARTH           2
#define MARS            3
#define JUPITER         4 
#define SATURN          5
#define URANUS          6
#define NEPTUNE         7
#define PLUTO           8

/*
 * Calculate the position of a planet for a given
 * julian date. Returns the right ascension and
 * declination in radians. planet is the planet number
 * listed above.
 */
void
PlanetPos(double *ra,
          double *dec,
          int planet,
          double jday);

/*
 * Calculate the rising and setting times for the moon.
 *
 * Times in UT.  If the object does not rise/set at this
 * latitude, returns -1, and UTr and UTs are not set.
 *
 * lat and lng are the observer's position in radians.
 */
int
PlanetRiseSet(int planet,
              int year,
              int month,
              int day,
              double lat,
              double lng,
              double *UTr,
              double *UTs);

typedef struct _AstroRecord
{
    char description[32]; // max 32 chars
    short raH, raM, raS;
    short decH, decM, decS;
    char magnitude[4]; // max 4 chars
    char size[8]; // max 8 chars
} AstroRecord;


#endif    // AstroLib_h

